from __future__ import annotations

from math import inf
from typing import TYPE_CHECKING
from unittest.mock import ANY

from pipdeptree.__main__ import main

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_json_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_json")
    main(["--json"])
    render.assert_called_once_with(ANY)


def test_json_tree_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_json_tree")
    main(["--json-tree"])
    render.assert_called_once_with(ANY)


def test_mermaid_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_mermaid")
    main(["--mermaid"])
    render.assert_called_once_with(ANY)


def test_grahpviz_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_graphviz")
    main(["--graph-output", "dot"])
    render.assert_called_once_with(ANY, output_format="dot", reverse=False)


def test_text_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_text")
    main([])
    render.assert_called_once_with(ANY, encoding="utf-8", max_depth=inf, list_all=False, include_license=False)


def test_freeze_routing(mocker: MockerFixture) -> None:
    render = mocker.patch("pipdeptree._render.render_freeze")
    main(["--freeze"])
    render.assert_called_once_with(ANY, max_depth=inf, list_all=False)



from __future__ import annotations

import locale
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pip._internal.models.direct_url import (
    DirectUrl,  # noqa: PLC2701
    DirectUrlValidationError,  # noqa: PLC2701
)
from pip._internal.utils.egg_link import egg_link_path_from_sys_path  # noqa: PLC2701
from pip._vendor.packaging.version import Version  # noqa: PLC2701

if TYPE_CHECKING:
    from importlib.metadata import Distribution


def dist_to_frozen_repr(dist: Distribution) -> str:
    """Return the frozen requirement repr of a `importlib.metadata.Distribution` object."""
    from pip._internal.operations.freeze import FrozenRequirement  # noqa: PLC0415, PLC2701

    adapter = PipBaseDistributionAdapter(dist)
    fr = FrozenRequirement.from_dist(adapter)  # type: ignore[arg-type]

    return str(fr).strip()


class PipBaseDistributionAdapter:
    """
    An adapter class for pip's `pip._internal.metadata.BaseDistribution` abstract class.

    It essentially wraps over an importlib.metadata.Distribution object and provides just enough fields/methods found in
    pip's `BaseDistribution` so that we can use `pip._internal.operations.freeze.FrozenRequirement.from_dist()`.

    :param dist: Represents an `importlib.metadata.Distribution` object.
    """

    DIRECT_URL_METADATA_NAME = "direct_url.json"

    def __init__(self, dist: Distribution) -> None:
        self._dist = dist
        self._raw_name = dist.metadata["Name"]
        self._version = Version(dist.version)

    @property
    def raw_name(self) -> str | Any:
        return self._raw_name

    @property
    def version(self) -> Version:
        return self._version

    @property
    def editable(self) -> bool:
        return self.editable_project_location is not None

    @property
    def direct_url(self) -> DirectUrl | None:
        result = None
        json_str = self._dist.read_text(self.DIRECT_URL_METADATA_NAME)
        try:
            if json_str:
                result = DirectUrl.from_json(json_str)
        except (
            UnicodeDecodeError,
            JSONDecodeError,
            DirectUrlValidationError,
        ):
            return result
        return result

    @property
    def editable_project_location(self) -> str | None:
        direct_url = self.direct_url
        if direct_url and direct_url.is_local_editable():
            from pip._internal.utils.urls import url_to_path  # noqa: PLC2701, PLC0415

            return url_to_path(direct_url.url)

        result = None
        egg_link_path = egg_link_path_from_sys_path(self.raw_name)
        if egg_link_path:
            with Path(egg_link_path).open("r", encoding=locale.getpreferredencoding(False)) as f:  # noqa: FBT003
                result = f.readline().rstrip()
        return result


__all__ = ["dist_to_frozen_repr"]



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any

from .text import get_top_level_nodes

if TYPE_CHECKING:
    from pipdeptree._models.dag import PackageDAG
    from pipdeptree._models.package import DistPackage, ReqPackage


def render_freeze(tree: PackageDAG, *, max_depth: float, list_all: bool = True) -> None:
    nodes = get_top_level_nodes(tree, list_all=list_all)

    def aux(
        node: DistPackage | ReqPackage,
        parent: DistPackage | ReqPackage | None = None,
        indent: int = 0,
        cur_chain: list[str] | None = None,
        depth: int = 0,
    ) -> list[Any]:
        cur_chain = cur_chain or []
        node_str = node.render(parent, frozen=True)
        if parent:
            prefix = " " * indent
            node_str = prefix + node_str
        result = [node_str]
        children = [
            aux(c, node, indent=indent + 2, cur_chain=[*cur_chain, c.project_name], depth=depth + 1)
            for c in tree.get_children(node.key)
            if c.project_name not in cur_chain and depth + 1 <= max_depth
        ]
        result += list(chain.from_iterable(children))
        return result

    lines = chain.from_iterable([aux(p) for p in nodes])
    print("\n".join(lines))  # noqa: T201


__all__ = [
    "render_freeze",
]



from __future__ import annotations

from pathlib import Path
from subprocess import CompletedProcess  # noqa: S404
from typing import TYPE_CHECKING

import pytest

from pipdeptree._detect_env import detect_active_interpreter

if TYPE_CHECKING:
    from pytest_mock import MockFixture


@pytest.mark.parametrize(("env_var"), ["VIRTUAL_ENV", "CONDA_PREFIX"])
def test_detect_active_interpreter_using_env_vars(tmp_path: Path, mocker: MockFixture, env_var: str) -> None:
    mocker.patch("pipdeptree._detect_env.os.environ", {env_var: str(tmp_path)})
    mocker.patch("pipdeptree._detect_env.Path.exists", return_value=True)

    actual_path = detect_active_interpreter()

    assert actual_path.startswith(str(tmp_path))


def test_detect_active_interpreter_poetry(tmp_path: Path, mocker: MockFixture) -> None:
    faked_result = CompletedProcess("", 0, stdout=str(tmp_path))
    mocker.patch("pipdeptree._detect_env.subprocess.run", return_value=faked_result)
    mocker.patch("pipdeptree._detect_env.os.environ", {})

    actual_path = detect_active_interpreter()

    assert str(tmp_path) == actual_path


def test_detect_active_interpreter_non_supported_python_implementation(
    tmp_path: Path,
    mocker: MockFixture,
) -> None:
    mocker.patch("pipdeptree._detect_env.os.environ", {"VIRTUAL_ENV": str(tmp_path)})
    mocker.patch("pipdeptree._detect_env.Path.exists", return_value=True)
    mocker.patch("pipdeptree._detect_env.platform.python_implementation", return_value="NotSupportedPythonImpl")

    with pytest.raises(SystemExit):
        detect_active_interpreter()


def test_detect_active_interpreter_non_existent_path(
    mocker: MockFixture,
) -> None:
    fake_path = str(Path(*("i", "dont", "exist")))
    mocker.patch("pipdeptree._detect_env.os.environ", {"VIRTUAL_ENV": fake_path})

    with pytest.raises(SystemExit):
        detect_active_interpreter()



from __future__ import annotations

import locale
from pathlib import Path
from random import shuffle
from typing import TYPE_CHECKING, Callable
from unittest.mock import Mock

import pytest

from pipdeptree._models import PackageDAG

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.our_types import MockGraph


@pytest.fixture(scope="session")
def mock_pkgs() -> Callable[[MockGraph], Iterator[Mock]]:
    def func(simple_graph: MockGraph) -> Iterator[Mock]:
        for node, children in simple_graph.items():
            nk, nv = node
            m = Mock(metadata={"Name": nk}, version=nv)
            reqs = []
            for ck, cv in children:
                r = ck
                for item in cv:
                    if item:
                        rs, rv = item
                        r = r + rs + rv
                    if item != cv[-1]:
                        r += ","
                reqs.append(r)
            m.requires = reqs
            yield m

    return func


@pytest.fixture
def example_dag(mock_pkgs: Callable[[MockGraph], Iterator[Mock]]) -> PackageDAG:
    packages: MockGraph = {
        ("a", "3.4.0"): [("b", [(">=", "2.0.0")]), ("c", [(">=", "5.7.1")])],
        ("b", "2.3.1"): [("d", [(">=", "2.30"), ("<", "2.42")])],
        ("c", "5.10.0"): [("d", [(">=", "2.30")]), ("e", [(">=", "0.12.1")])],
        ("d", "2.35"): [("e", [(">=", "0.9.0")])],
        ("e", "0.12.1"): [],
        ("f", "3.1"): [("b", [(">=", "2.1.0")])],
        ("g", "6.8.3rc1"): [("e", [(">=", "0.9.0")]), ("f", [(">=", "3.0.0")])],
    }
    return PackageDAG.from_pkgs(list(mock_pkgs(packages)))


@pytest.fixture
def randomized_example_dag(example_dag: PackageDAG) -> PackageDAG:
    """Returns a copy of the package tree fixture with dependencies in randomized order."""
    # Extract the dependency graph from the package tree and randomize it.
    randomized_graph = {}
    randomized_nodes = list(example_dag._obj.keys())  # noqa: SLF001
    shuffle(randomized_nodes)
    for node in randomized_nodes:
        edges = example_dag._obj[node].copy()  # noqa: SLF001
        shuffle(edges)
        randomized_graph[node] = edges
    assert set(randomized_graph) == set(example_dag._obj)  # noqa: SLF001

    # Create a randomized package tree.
    randomized_dag = PackageDAG(randomized_graph)
    assert len(example_dag) == len(randomized_dag)
    return randomized_dag


@pytest.fixture
def fake_dist(tmp_path: Path) -> Path:
    """Creates a fake site package (that you get using Path.parent) and a fake dist-info called bar-2.4.5.dist-info."""
    fake_site_pkgs = tmp_path / "site-packages"
    fake_dist_path = fake_site_pkgs / "bar-2.4.5.dist-info"
    fake_dist_path.mkdir(parents=True)
    fake_metadata = Path(fake_dist_path) / "METADATA"
    with fake_metadata.open("w", encoding=locale.getpreferredencoding(False)) as f:
        f.write("Metadata-Version: 2.3\nName: bar\nVersion: 2.4.5\n")

    return fake_dist_path


@pytest.fixture
def fake_dist_with_invalid_metadata(tmp_path: Path) -> Path:
    "Similar to `fake_dist()`, but creates an invalid METADATA file."
    fake_site_pkgs = tmp_path / "site-packages"
    fake_dist_path = fake_site_pkgs / "bar-2.4.5.dist-info"
    fake_dist_path.mkdir(parents=True)
    fake_metadata = Path(fake_dist_path) / "METADATA"
    fake_metadata.touch()
    return fake_dist_path



from __future__ import annotations

import sys
from platform import python_implementation
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest
import virtualenv

from pipdeptree.__main__ import main

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture


@pytest.fixture(scope="session")
def expected_venv_pkgs() -> frozenset[str]:
    implementation = python_implementation()
    if implementation == "CPython":  # pragma: cpython cover
        expected = {"pip", "setuptools"}
    elif implementation == "PyPy":  # pragma: pypy cover
        expected = {"cffi", "greenlet", "pip", "hpy", "readline", "setuptools"}
        if sys.version_info >= (3, 10):  # pragma: >=3.10 cover
            expected -= {"readline"}
    else:  # pragma: no cover
        raise ValueError(implementation)
    if sys.version_info >= (3, 12):  # pragma: >=3.12 cover
        expected -= {"setuptools"}

    return frozenset(expected)


@pytest.mark.parametrize("args_joined", [True, False])
def test_custom_interpreter(
    tmp_path: Path,
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    capfd: pytest.CaptureFixture[str],
    args_joined: bool,
    expected_venv_pkgs: frozenset[str],
) -> None:
    # Delete $PYTHONPATH so that it cannot be passed to the custom interpreter process (since we don't know what
    # distribution metadata to expect when it's used).
    monkeypatch.delenv("PYTHONPATH", False)

    monkeypatch.chdir(tmp_path)
    result = virtualenv.cli_run([str(tmp_path / "venv"), "--activators", ""])
    py = str(result.creator.exe.relative_to(tmp_path))
    cmd = ["", f"--python={result.creator.exe}"] if args_joined else ["", "--python", py]
    cmd += ["--all", "--depth", "0"]
    mocker.patch("pipdeptree._discovery.sys.argv", cmd)
    main()
    out, _ = capfd.readouterr()
    found = {i.split("==")[0] for i in out.splitlines()}

    assert expected_venv_pkgs == found, out


def test_custom_interpreter_with_local_only(
    tmp_path: Path,
    mocker: MockerFixture,
    capfd: pytest.CaptureFixture[str],
) -> None:
    venv_path = str(tmp_path / "venv")
    result = virtualenv.cli_run([venv_path, "--system-site-packages", "--activators", ""])

    cmd = ["", f"--python={result.creator.exe}", "--local-only"]
    mocker.patch("pipdeptree._discovery.sys.prefix", venv_path)
    mocker.patch("pipdeptree._discovery.sys.argv", cmd)
    main()
    out, _ = capfd.readouterr()
    found = {i.split("==")[0] for i in out.splitlines()}
    expected = {"pip", "setuptools"}
    if sys.version_info >= (3, 12):  # pragma: >=3.12 cover
        expected -= {"setuptools"}
    assert expected == found, out


def test_custom_interpreter_with_user_only(
    tmp_path: Path, mocker: MockerFixture, capfd: pytest.CaptureFixture[str]
) -> None:
    # ensures there is no output when --user-only and --python are passed

    venv_path = str(tmp_path / "venv")
    result = virtualenv.cli_run([venv_path, "--activators", ""])

    cmd = ["", f"--python={result.creator.exe}", "--user-only"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)
    main()
    out, err = capfd.readouterr()
    assert not err

    # Here we expect 1 element because print() adds a newline.
    found = out.splitlines()
    assert len(found) == 1
    assert not found[0]


def test_custom_interpreter_with_user_only_and_system_site_pkgs_enabled(
    tmp_path: Path,
    fake_dist: Path,
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    capfd: pytest.CaptureFixture[str],
) -> None:
    # ensures that we provide user site metadata when --user-only and --python are passed and the custom interpreter has
    # system site packages enabled

    # Make a fake user site directory since we don't know what to expect from the real one.
    fake_user_site = str(fake_dist.parent)
    mocker.patch("pipdeptree._discovery.site.getusersitepackages", Mock(return_value=fake_user_site))

    # Create a temporary virtual environment.
    venv_path = str(tmp_path / "venv")
    result = virtualenv.cli_run([venv_path, "--activators", ""])

    # Use $PYTHONPATH to add the fake user site into the custom interpreter's environment so that it will include it in
    # its sys.path.
    monkeypatch.setenv("PYTHONPATH", str(fake_user_site))

    cmd = ["", f"--python={result.creator.exe}", "--user-only"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)
    main()

    out, err = capfd.readouterr()
    assert not err
    found = {i.split("==")[0] for i in out.splitlines()}
    expected = {"bar"}

    assert expected == found


def test_custom_interpreter_ensure_pythonpath_envar_is_honored(
    tmp_path: Path,
    mocker: MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    capfd: pytest.CaptureFixture[str],
    expected_venv_pkgs: frozenset[str],
) -> None:
    # ensures that we honor $PYTHONPATH when passing it to the custom interpreter process
    venv_path = str(tmp_path / "venv")
    result = virtualenv.cli_run([venv_path, "--activators", ""])

    another_path = tmp_path / "another-path"
    fake_dist = another_path / "foo-1.2.3.dist-info"
    fake_dist.mkdir(parents=True)
    fake_metadata = fake_dist / "METADATA"
    with fake_metadata.open("w") as f:
        f.write("Metadata-Version: 2.3\nName: foo\nVersion: 1.2.3\n")
    cmd = ["", f"--python={result.creator.exe}", "--all", "--depth", "0"]
    mocker.patch("pipdeptree._discovery.sys.argv", cmd)
    monkeypatch.setenv("PYTHONPATH", str(another_path))
    main()
    out, _ = capfd.readouterr()
    found = {i.split("==")[0] for i in out.splitlines()}
    assert {*expected_venv_pkgs, "foo"} == found, out



from __future__ import annotations

from math import inf
from typing import TYPE_CHECKING
from unittest.mock import PropertyMock

import pytest

from pipdeptree._freeze import PipBaseDistributionAdapter
from pipdeptree._render.freeze import render_freeze

if TYPE_CHECKING:
    from pipdeptree._models.dag import PackageDAG


@pytest.fixture
def patch_pip_adapter(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Patches `PipBaseDistributionAdapter` such that `editable` returns `False` and `direct_url` returns `None`.

    This will have the pip API always return a frozen req in the "name==version" format.
    """
    monkeypatch.setattr(PipBaseDistributionAdapter, "editable", PropertyMock(return_value=False))
    monkeypatch.setattr(PipBaseDistributionAdapter, "direct_url", PropertyMock(return_value=None))


@pytest.mark.parametrize(
    ("list_all", "expected_output"),
    [
        (
            True,
            [
                "a==3.4.0",
                "  b==2.3.1",
                "    d==2.35",
                "      e==0.12.1",
                "  c==5.10.0",
                "    d==2.35",
                "      e==0.12.1",
                "    e==0.12.1",
                "b==2.3.1",
                "  d==2.35",
                "    e==0.12.1",
                "c==5.10.0",
                "  d==2.35",
                "    e==0.12.1",
                "  e==0.12.1",
                "d==2.35",
                "  e==0.12.1",
                "e==0.12.1",
                "f==3.1",
                "  b==2.3.1",
                "    d==2.35",
                "      e==0.12.1",
                "g==6.8.3rc1",
                "  e==0.12.1",
                "  f==3.1",
                "    b==2.3.1",
                "      d==2.35",
                "        e==0.12.1",
            ],
        ),
        (
            False,
            [
                "a==3.4.0",
                "  b==2.3.1",
                "    d==2.35",
                "      e==0.12.1",
                "  c==5.10.0",
                "    d==2.35",
                "      e==0.12.1",
                "    e==0.12.1",
                "g==6.8.3rc1",
                "  e==0.12.1",
                "  f==3.1",
                "    b==2.3.1",
                "      d==2.35",
                "        e==0.12.1",
            ],
        ),
    ],
)
@pytest.mark.usefixtures("patch_pip_adapter")
def test_render_freeze(
    example_dag: PackageDAG,
    capsys: pytest.CaptureFixture[str],
    list_all: bool,
    expected_output: list[str],
) -> None:
    render_freeze(example_dag, max_depth=inf, list_all=list_all)
    captured = capsys.readouterr()
    assert "\n".join(expected_output).strip() == captured.out.strip()


@pytest.mark.parametrize(
    ("depth", "expected_output"),
    [
        (
            0,
            [
                "a==3.4.0",
                "b==2.3.1",
                "c==5.10.0",
                "d==2.35",
                "e==0.12.1",
                "f==3.1",
                "g==6.8.3rc1",
            ],
        ),
        (
            2,
            [
                "a==3.4.0",
                "  b==2.3.1",
                "    d==2.35",
                "  c==5.10.0",
                "    d==2.35",
                "    e==0.12.1",
                "b==2.3.1",
                "  d==2.35",
                "    e==0.12.1",
                "c==5.10.0",
                "  d==2.35",
                "    e==0.12.1",
                "  e==0.12.1",
                "d==2.35",
                "  e==0.12.1",
                "e==0.12.1",
                "f==3.1",
                "  b==2.3.1",
                "    d==2.35",
                "g==6.8.3rc1",
                "  e==0.12.1",
                "  f==3.1",
                "    b==2.3.1",
            ],
        ),
    ],
)
@pytest.mark.usefixtures("patch_pip_adapter")
def test_render_freeze_given_depth(
    example_dag: PackageDAG,
    capsys: pytest.CaptureFixture[str],
    depth: int,
    expected_output: list[str],
) -> None:
    render_freeze(example_dag, max_depth=depth)
    captured = capsys.readouterr()
    assert "\n".join(expected_output).strip() == captured.out.strip()



"""The main entry point used for CLI."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from pipdeptree._cli import get_options
from pipdeptree._detect_env import detect_active_interpreter
from pipdeptree._discovery import InterpreterQueryError, get_installed_distributions
from pipdeptree._models import PackageDAG
from pipdeptree._models.dag import IncludeExcludeOverlapError, IncludePatternNotFoundError
from pipdeptree._render import render
from pipdeptree._validate import validate
from pipdeptree._warning import WarningPrinter, WarningType, get_warning_printer

if TYPE_CHECKING:
    from collections.abc import Sequence


def main(args: Sequence[str] | None = None) -> int | None:
    """CLI - The main function called as entry point."""
    options = get_options(args)

    # Warnings are only enabled when using text output.
    is_text_output = not any([options.json, options.json_tree, options.output_format])
    if not is_text_output:
        options.warn = WarningType.SILENCE
    warning_printer = get_warning_printer()
    warning_printer.warning_type = options.warn

    if options.python == "auto":
        resolved_path = detect_active_interpreter()
        options.python = resolved_path
        print(f"(resolved python: {resolved_path})", file=sys.stderr)  # noqa: T201

    try:
        pkgs = get_installed_distributions(
            interpreter=options.python,
            supplied_paths=options.path or None,
            local_only=options.local_only,
            user_only=options.user_only,
        )
    except InterpreterQueryError as e:
        print(f"Failed to query custom interpreter: {e}", file=sys.stderr)  # noqa: T201
        return 1

    tree = PackageDAG.from_pkgs(pkgs)

    validate(tree)

    # Reverse the tree (if applicable) before filtering, thus ensuring, that the filter will be applied on ReverseTree
    if options.reverse:
        tree = tree.reverse()

    include = options.packages.split(",") if options.packages else None
    exclude = set(options.exclude.split(",")) if options.exclude else None

    if include is not None or exclude is not None:
        try:
            tree = tree.filter_nodes(include, exclude, exclude_deps=options.exclude_dependencies)
        except IncludeExcludeOverlapError:
            print("Cannot have --packages and --exclude contain the same entries", file=sys.stderr)  # noqa: T201
            return 1
        except IncludePatternNotFoundError as e:
            if warning_printer.should_warn():
                warning_printer.print_single_line(str(e))
            return _determine_return_code(warning_printer)

    render(options, tree)

    return _determine_return_code(warning_printer)


def _determine_return_code(warning_printer: WarningPrinter) -> int:
    return 1 if warning_printer.has_warned_with_failure() else 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from unittest.mock import Mock

from pipdeptree._freeze import dist_to_frozen_repr


def test_dist_to_frozen_repr() -> None:
    foo = Mock(metadata={"Name": "foo"}, version="20.4.1")
    foo.read_text = Mock(return_value=None)
    expected = "foo==20.4.1"
    assert dist_to_frozen_repr(foo) == expected



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any, Callable

import pytest

from pipdeptree._models import DistPackage, PackageDAG, ReqPackage, ReversedPackageDAG
from pipdeptree._models.dag import IncludeExcludeOverlapError, IncludePatternNotFoundError

if TYPE_CHECKING:
    from collections.abc import Iterator
    from unittest.mock import Mock

    from tests.our_types import MockGraph


def test_package_dag_get_node_as_parent(example_dag: PackageDAG) -> None:
    node = example_dag.get_node_as_parent("b")
    assert node is not None
    assert node.key == "b"
    node = example_dag.get_node_as_parent("c")
    assert node is not None
    assert node.key == "c"


@pytest.fixture(scope="session")
def t_fnmatch(mock_pkgs: Callable[[MockGraph], Iterator[Mock]]) -> PackageDAG:
    graph: MockGraph = {
        ("a-a", "1"): [("a-b", []), ("a-c", [])],
        ("a-b", "1"): [("a-c", [])],
        ("b-a", "1"): [("b-b", [])],
        ("b-b", "1"): [("a-b", [])],
    }
    return PackageDAG.from_pkgs(list(mock_pkgs(graph)))


def dag_to_dict(g: PackageDAG) -> dict[str, list[str]]:
    return {k.key: [v.key for v in vs] for k, vs in g._obj.items()}  # noqa: SLF001


def test_package_dag_filter_fnmatch_include_a(t_fnmatch: PackageDAG) -> None:
    # test include for a-*in the result we got only a-* nodes
    graph = dag_to_dict(t_fnmatch.filter_nodes(["a-*"], None))
    assert graph == {"a-a": ["a-b", "a-c"], "a-b": ["a-c"]}


def test_package_dag_filter_fnmatch_include_b(t_fnmatch: PackageDAG) -> None:
    # test include for b-*, which has a-b and a-c in tree, but not a-a
    # in the result we got the b-* nodes plus the a-b node as child in the tree
    graph = dag_to_dict(t_fnmatch.filter_nodes(["b-*"], None))
    assert graph == {"b-a": ["b-b"], "b-b": ["a-b"], "a-b": ["a-c"]}


def test_package_dag_filter_fnmatch_exclude_c(t_fnmatch: PackageDAG) -> None:
    # test exclude for b-* in the result we got only a-* nodes
    graph = dag_to_dict(t_fnmatch.filter_nodes(None, {"b-*"}))
    assert graph == {"a-a": ["a-b", "a-c"], "a-b": ["a-c"]}


def test_package_dag_filter_fnmatch_exclude_a(t_fnmatch: PackageDAG) -> None:
    # test exclude for a-* in the result we got only b-* nodes
    graph = dag_to_dict(t_fnmatch.filter_nodes(None, {"a-*"}))
    assert graph == {"b-a": ["b-b"], "b-b": []}


def test_package_dag_filter_include_exclude_normal(t_fnmatch: PackageDAG) -> None:
    graph = dag_to_dict(t_fnmatch.filter_nodes(["a-*"], {"a-a"}))
    assert graph == {"a-b": ["a-c"]}


def test_package_dag_filter_include_exclude_overlap(t_fnmatch: PackageDAG) -> None:
    with pytest.raises(IncludeExcludeOverlapError):
        t_fnmatch.filter_nodes(["a-a", "a-b"], {"a-b"})


def test_package_dag_filter_include_nonexistent_packages(t_fnmatch: PackageDAG) -> None:
    with pytest.raises(IncludePatternNotFoundError, match="No packages matched using the following patterns: x, y, z"):
        t_fnmatch.filter_nodes(["x", "y", "z"], None)


def test_package_dag_filter_packages_uses_pep503normalize(
    mock_pkgs: Callable[[MockGraph], Iterator[Mock]],
) -> None:
    graph: MockGraph = {
        ("Pie.Pie", "1"): [],
    }
    pkgs = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    pkgs = pkgs.filter_nodes(["Pie.Pie"], None)
    assert len(pkgs) == 1
    assert pkgs.get_node_as_parent("pie-pie") is not None

    pkgs = pkgs.filter_nodes(None, {"Pie.Pie"})
    assert len(pkgs) == 0


@pytest.mark.parametrize(
    ("graph", "exclude", "expected"),
    [
        pytest.param(
            {
                ("a", "1.0.0"): [("b", []), ("c", []), ("d", []), ("e", [])],
                ("b", "2.0.0"): [("x", []), ("y", []), ("z", [])],
                ("c", "3.0.0"): [("d", [])],
                ("d", "4.0.0"): [],
                ("e", "5.0.0"): [("y", [])],
                ("f", "6.0.0"): [("d", [])],  # Make sure that "d" is not removed since "f" is not excluded
                ("g", "7.0.0"): [],
                ("x", "8.0.0"): [],
                ("y", "9.0.0"): [],
                ("z", "10.0.0"): [],
            },
            {"a"},
            {"d", "f", "g"},
            id="with-dependencies",
        ),
        pytest.param(
            {
                ("acorn", "1.0.0"): [("b", []), ("c", []), ("d", []), ("e", [])],
                ("b", "2.0.0"): [("x", []), ("y", []), ("z", [])],
                ("c", "3.0.0"): [("d", [])],
                ("d", "4.0.0"): [],
                ("e", "5.0.0"): [("y", [])],
                ("f", "6.0.0"): [("d", [])],
                ("g", "7.0.0"): [],
                ("x", "8.0.0"): [],
                ("y", "9.0.0"): [],
                ("z", "10.0.0"): [],
            },
            {"a*", "g"},
            {"d", "f"},
            id="with-multiple-excludes-and-with-patterns",
        ),
        pytest.param(
            {
                ("b", "2.0.0"): [("x", []), ("y", []), ("z", [])],
                ("x", "8.0.0"): [],
                ("y", "9.0.0"): [],
                ("z", "10.0.0"): [],
            },
            {"dontexist"},
            {"b", "x", "y", "z"},
            id="with-non-existent-exclude",
        ),
        pytest.param(
            {
                ("a", "1.0.0"): [("b", [])],
                ("b", "2.0.0"): [("c", [])],
                ("c", "3.0.0"): [("d", [])],
                ("d", "4.0.0"): [("e", [])],
                ("e", "5.0.0"): [("f", [])],
                ("f", "6.0.0"): [("g", [])],
                ("g", "7.0.0"): [("x", [])],
                ("x", "8.0.0"): [("y", [])],
                ("y", "9.0.0"): [("z", [])],
                ("z", "10.0.0"): [],
            },
            {"c"},
            {"a", "b"},
            id="with-package-having-large-depth",
        ),
    ],
)
def test_package_dag_filter_packages_given_exclude_dependencies(
    mock_pkgs: Callable[[MockGraph], Iterator[Mock]], graph: MockGraph, exclude: set[str], expected: list[str]
) -> None:
    pkgs = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    pkgs = pkgs.filter_nodes(None, exclude, exclude_deps=True)

    assert len(pkgs) == len(expected)
    assert all(p.key in expected for p in pkgs)


def test_package_dag_reverse(example_dag: PackageDAG) -> None:
    def sort_map_values(m: dict[str, Any]) -> dict[str, Any]:
        return {k: sorted(v) for k, v in m.items()}

    t1 = example_dag.reverse()
    expected = {"a": [], "b": ["a", "f"], "c": ["a"], "d": ["b", "c"], "e": ["c", "d", "g"], "f": ["g"], "g": []}
    assert isinstance(t1, ReversedPackageDAG)
    assert sort_map_values(expected) == sort_map_values(dag_to_dict(t1))
    assert all(isinstance(k, ReqPackage) for k in t1)
    assert all(isinstance(v, DistPackage) for v in chain.from_iterable(t1.values()))

    # testing reversal of ReversedPackageDAG instance
    expected = {"a": ["b", "c"], "b": ["d"], "c": ["d", "e"], "d": ["e"], "e": [], "f": ["b"], "g": ["e", "f"]}
    t2 = t1.reverse()
    assert isinstance(t2, PackageDAG)
    assert sort_map_values(expected) == sort_map_values(dag_to_dict(t2))
    assert all(isinstance(k, DistPackage) for k in t2)
    assert all(isinstance(v, ReqPackage) for v in chain.from_iterable(t2.values()))


def test_package_dag_from_pkgs(mock_pkgs: Callable[[MockGraph], Iterator[Mock]]) -> None:
    # when pip's _vendor.packaging.requirements.Requirement's requires() gives a lowercased package name but the actual
    # package name in PyPI is mixed case, expect the mixed case version

    graph: MockGraph = {
        ("examplePy", "1.2.3"): [("hellopy", [(">=", "2.0.0")])],
        ("HelloPy", "2.2.0"): [],
    }
    package_dag = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    parent_key = "examplepy"
    c = package_dag.get_children(parent_key)
    assert len(c) == 1
    assert c[0].project_name == "HelloPy"


def test_package_dag_from_pkgs_uses_pep503normalize(mock_pkgs: Callable[[MockGraph], Iterator[Mock]]) -> None:
    # ensure that requirement gets matched with a dists even when it's key needs pep503 normalization to match

    graph: MockGraph = {
        ("parent-package", "1.2.3"): [("flufl.lock", [(">=", "2.0.0")])],
        ("flufl-lock", "2.2.0"): [],
    }
    package_dag = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    parent_key = "parent-package"
    c = package_dag.get_children(parent_key)
    assert c[0].dist
    assert c[0].key == "flufl-lock"


def test_package_from_pkgs_given_invalid_requirements(
    mock_pkgs: Callable[[MockGraph], Iterator[Mock]], capfd: pytest.CaptureFixture[str]
) -> None:
    graph: MockGraph = {
        ("a-package", "1.2.3"): [("BAD**requirement", [(">=", "2.0.0")])],
    }
    package_dag = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    assert len(package_dag) == 1
    out, err = capfd.readouterr()
    assert not out
    assert err == (
        "Warning!!! Invalid requirement strings found for the following distributions:\na-package\n  "
        'Skipping "BAD**requirement>=2.0.0"\n------------------------------------------------------------------------\n'
    )



from __future__ import annotations

from textwrap import dedent, indent
from typing import TYPE_CHECKING, Callable

from pipdeptree._models import PackageDAG
from pipdeptree._render.mermaid import render_mermaid

if TYPE_CHECKING:
    from collections.abc import Iterator
    from unittest.mock import Mock

    from tests.our_types import MockGraph


def test_render_mermaid(example_dag: PackageDAG, randomized_example_dag: PackageDAG) -> None:
    """Check both the sorted and randomized package tree produces the same sorted Mermaid output.

    Rendering a reverse dependency tree should produce the same set of nodes. Edges should have the same version spec
    label, but be resorted after swapping node positions.

    `See how this renders <https://mermaid.ink/img/pako:eNp9kcluwjAURX_FeutgeUhCErWs-IN21boL4yFEzYAyqKWIf-9LCISyqFf2ufe-QT6BaayDDHzZfJm9bnvyulU1wWNK3XVb50lVdF1R56Tr2-bTrazu0NfqY0aii1O_K9BK1ZKGlCn4uNAd0h1SQSXlN2qQGqQR5ezObBHbizm6QYfQIWSUi7sSHrGf2i0sR5Yji2lCZWsWQZPViijYPAs69cPnhuwetIiux1qTZubpl5xkwZOgoZgNdl7karhON4nuQRzTf3N2yaW3geaYX2L8cdj8n9xNk3dLegigcm2lC4v_exqdCvq9q5yCDK_WeT2UvQJVn9Gqh755OdYGsr4dXADDwerebQudt7qCzOuyQ3rQ9VvTVFcTPiE7wTdkgkVUSiHjlLOERUkYB3AcMeXrhKUp53GYcJ6KcwA_UwVGo_MvVqym_A?type=png)](https://mermaid.live/edit#pako:eNp9kcluwjAURX_FeutgeUhCErWs-IN21boL4yFEzYAyqKWIf-9LCISyqFf2ufe-QT6BaayDDHzZfJm9bnvyulU1wWNK3XVb50lVdF1R56Tr2-bTrazu0NfqY0aii1O_K9BK1ZKGlCn4uNAd0h1SQSXlN2qQGqQR5ezObBHbizm6QYfQIWSUi7sSHrGf2i0sR5Yji2lCZWsWQZPViijYPAs69cPnhuwetIiux1qTZubpl5xkwZOgoZgNdl7karhON4nuQRzTf3N2yaW3geaYX2L8cdj8n9xNk3dLegigcm2lC4v_exqdCvq9q5yCDK_WeT2UvQJVn9Gqh755OdYGsr4dXADDwerebQudt7qCzOuyQ3rQ9VvTVFcTPiE7wTdkgkVUSiHjlLOERUkYB3AcMeXrhKUp53GYcJ6KcwA_UwVGo_MvVqym_A>`_.

    """

    nodes = dedent(
        """\
        flowchart TD
            classDef missing stroke-dasharray: 5
            a["a\\n3.4.0"]
            b["b\\n2.3.1"]
            c["c\\n5.10.0"]
            d["d\\n2.35"]
            e["e\\n0.12.1"]
            f["f\\n3.1"]
            g["g\\n6.8.3rc1"]
        """,
    )
    dependency_edges = indent(
        dedent(
            """\
            a -- ">=2.0.0" --> b
            a -- ">=5.7.1" --> c
            b -- ">=2.30,<2.42" --> d
            c -- ">=0.12.1" --> e
            c -- ">=2.30" --> d
            d -- ">=0.9.0" --> e
            f -- ">=2.1.0" --> b
            g -- ">=0.9.0" --> e
            g -- ">=3.0.0" --> f
        """,
        ),
        " " * 4,
    ).rstrip()
    reverse_dependency_edges = indent(
        dedent(
            """\
            b -- ">=2.0.0" --> a
            b -- ">=2.1.0" --> f
            c -- ">=5.7.1" --> a
            d -- ">=2.30" --> c
            d -- ">=2.30,<2.42" --> b
            e -- ">=0.12.1" --> c
            e -- ">=0.9.0" --> d
            e -- ">=0.9.0" --> g
            f -- ">=3.0.0" --> g
        """,
        ),
        " " * 4,
    ).rstrip()

    for package_tree in (example_dag, randomized_example_dag):
        output = render_mermaid(package_tree)
        assert output.rstrip() == nodes + dependency_edges
        reversed_output = render_mermaid(package_tree.reverse())
        assert reversed_output.rstrip() == nodes + reverse_dependency_edges


def test_mermaid_reserved_ids(mock_pkgs: Callable[[MockGraph], Iterator[Mock]]) -> None:
    graph = {("click", "3.4.0"): [("click-extra", [(">=", "2.0.0")])]}
    package_tree = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    output = render_mermaid(package_tree)
    assert output == dedent(
        """\
        flowchart TD
            classDef missing stroke-dasharray: 5
            click-extra["click-extra\\n(missing)"]:::missing
            click_0["click\\n3.4.0"]
            click_0 -.-> click-extra
        """,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

from pipdeptree._models.dag import PackageDAG
from pipdeptree._render.json_tree import render_json_tree

if TYPE_CHECKING:
    from collections.abc import Iterator
    from unittest.mock import Mock

    from tests.our_types import MockGraph


@pytest.mark.parametrize(
    ("version_spec_tuple", "expected_version_spec"),
    [
        pytest.param((), "Any"),
        pytest.param((">=", "2.0.0"), ">=2.0.0"),
    ],
)
def test_json_tree_given_req_package_with_version_spec(
    mock_pkgs: Callable[[MockGraph], Iterator[Mock]],
    version_spec_tuple: tuple[str, str],
    expected_version_spec: str,
) -> None:
    graph: MockGraph = {
        ("a", "1.2.3"): [("b", [version_spec_tuple])],
        ("b", "2.2.0"): [],
    }
    package_dag = PackageDAG.from_pkgs(list(mock_pkgs(graph)))
    json_tree_str = render_json_tree(package_dag)
    assert json_tree_str.find(expected_version_spec) != -1



from __future__ import annotations

from typing import TYPE_CHECKING

from pipdeptree._warning import WarningPrinter, WarningType

if TYPE_CHECKING:
    import pytest


def test_warning_printer_print_single_line(capsys: pytest.CaptureFixture[str]) -> None:
    # Use WarningType.FAIL so that we can be able to test to see if WarningPrinter remembers it has warned before.
    warning_printer = WarningPrinter(WarningType.FAIL)
    warning_printer.print_single_line("test")
    assert warning_printer.has_warned_with_failure()
    out, err = capsys.readouterr()
    assert len(out) == 0
    assert err == "test\n"



from __future__ import annotations

import sys
from importlib import metadata
from subprocess import CompletedProcess, check_call  # noqa: S404
from typing import TYPE_CHECKING

import pytest

from pipdeptree.__main__ import main

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockFixture


def test_main() -> None:
    check_call([sys.executable, "-m", "pipdeptree", "--help"])


def test_console_script() -> None:
    try:
        dist = metadata.distribution("pipdeptree")
    except Exception as e:  # noqa: BLE001 # pragma: no cover
        pytest.fail(f"Unexpected error when retrieving pipdeptree metadata: {e}")

    entry_points = dist.entry_points
    assert len(entry_points) == 1

    if sys.version_info >= (3, 11):  # pragma: >=3.11
        entry_point = entry_points["pipdeptree"]
    else:
        entry_point = entry_points[0]

    try:
        pipdeptree = entry_point.load()
    except Exception as e:  # noqa: BLE001 # pragma: no cover
        pytest.fail(f"Unexpected error: {e}")

    with pytest.raises(SystemExit, match="0"):
        pipdeptree(["", "--help"])


def test_main_log_resolved(tmp_path: Path, mocker: MockFixture, capsys: pytest.CaptureFixture[str]) -> None:
    mocker.patch("sys.argv", ["", "--python", "auto"])
    mocker.patch("pipdeptree.__main__.detect_active_interpreter", return_value=str(tmp_path))
    mock_subprocess_run = mocker.patch("subprocess.run")
    valid_sys_path = str([str(tmp_path)])
    mock_subprocess_run.return_value = CompletedProcess(
        args=["python", "-c", "import sys; print(sys.path)"],
        returncode=0,
        stdout=valid_sys_path,
        stderr="",
    )

    main()

    captured = capsys.readouterr()
    assert captured.err.startswith(f"(resolved python: {tmp_path!s}")


def test_main_include_and_exclude_overlap(mocker: MockFixture, capsys: pytest.CaptureFixture[str]) -> None:
    cmd = ["", "--packages", "a,b,c", "--exclude", "a"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)

    main()

    captured = capsys.readouterr()
    assert "Cannot have --packages and --exclude contain the same entries" in captured.err



from __future__ import annotations

from importlib.metadata import PackageNotFoundError
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock, Mock

import pytest
from packaging.specifiers import SpecifierSet

from pipdeptree._models import DistPackage, ReqPackage
from pipdeptree._models.package import Package

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def sort_map_values(m: dict[str, Any]) -> dict[str, Any]:
    return {k: sorted(v) for k, v in m.items()}


def test_guess_version_setuptools(mocker: MockerFixture) -> None:
    mocker.patch("pipdeptree._models.package.version", side_effect=PackageNotFoundError)
    r = MagicMock()
    r.name = "setuptools"
    result = ReqPackage(r).installed_version
    assert result == "?"


def test_package_as_frozen_repr(mocker: MockerFixture) -> None:
    foo = Mock(metadata={"Name": "foo"}, version="1.2.3")
    dp = DistPackage(foo)
    expected = "test"
    mocker.patch("pipdeptree._models.package.dist_to_frozen_repr", Mock(return_value=expected))
    assert Package.as_frozen_repr(dp.unwrap()) == expected


def test_dist_package_requires() -> None:
    foo = Mock(
        metadata={"Name": "foo"},
        requires=["bar", "baz >=2.7.2"],
    )
    dp = DistPackage(foo)
    reqs = list(dp.requires())
    assert len(reqs) == 2


def test_dist_package_requires_with_environment_markers_that_eval_to_false() -> None:
    foo = Mock(
        metadata={"Name": "foo"},
        requires=['foo ; sys_platform == "NoexistOS"', "bar >=2.7.2 ; extra == 'testing'"],
    )
    dp = DistPackage(foo)
    reqs = list(dp.requires())
    assert len(reqs) == 0


def test_dist_package_render_as_root() -> None:
    foo = Mock(metadata={"Name": "foo"}, version="20.4.1")
    dp = DistPackage(foo)
    assert dp.render_as_root(frozen=False) == "foo==20.4.1"


def test_dist_package_render_as_branch() -> None:
    foo = Mock(metadata={"Name": "foo"}, version="20.4.1")
    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    bar_req = MagicMock(version="4.1.0", specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    dp = DistPackage(foo).as_parent_of(rp)
    assert dp.render_as_branch(frozen=False) == "foo==20.4.1 [requires: bar>=4.0]"


def test_dist_package_render_as_root_with_frozen(mocker: MockerFixture) -> None:
    foo = Mock(metadata={"Name": "foo"}, version="1.2.3")
    dp = DistPackage(foo)
    expected = "test"
    mocker.patch("pipdeptree._models.package.dist_to_frozen_repr", Mock(return_value=expected))
    assert dp.render_as_root(frozen=True) == expected


def test_dist_package_as_parent_of() -> None:
    foo = Mock(metadata={"Name": "foo"}, version="20.4.1")
    dp = DistPackage(foo)
    assert dp.req is None

    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    bar_req = MagicMock(version="4.1.0", specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    dp1 = dp.as_parent_of(rp)
    assert dp1._obj == dp._obj  # noqa: SLF001
    assert dp1.req is rp

    dp2 = dp.as_parent_of(None)
    assert dp2 is dp


def test_dist_package_as_dict() -> None:
    foo = Mock(metadata={"Name": "foo"}, version="1.3.2b1")
    dp = DistPackage(foo)
    result = dp.as_dict()
    expected = {"key": "foo", "package_name": "foo", "installed_version": "1.3.2b1"}
    assert expected == result


@pytest.mark.parametrize(
    ("mocked_metadata", "expected_output"),
    [
        pytest.param(
            Mock(get=Mock(return_value=None), get_all=Mock(return_value=[])),
            Package.UNKNOWN_LICENSE_STR,
            id="no-license",
        ),
        pytest.param(
            Mock(
                get=Mock(return_value=None),
                get_all=Mock(
                    return_value=[
                        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
                        "Operating System :: OS Independent",
                    ]
                ),
            ),
            "(GNU General Public License v2 (GPLv2))",
            id="one-license-with-one-non-license",
        ),
        pytest.param(
            Mock(
                get=Mock(return_value=None),
                get_all=Mock(
                    return_value=[
                        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
                        "License :: OSI Approved :: Apache Software License",
                    ]
                ),
            ),
            "(GNU General Public License v2 (GPLv2), Apache Software License)",
            id="more-than-one-license",
        ),
        pytest.param(
            Mock(get=Mock(return_value="MIT"), get_all=Mock(return_value=[])),
            "(MIT)",
            id="license-expression",
        ),
        pytest.param(
            Mock(
                get=Mock(return_value="MIT"),
                get_all=Mock(
                    return_value=[
                        "License :: OSI Approved :: MIT License",
                    ]
                ),
            ),
            "(MIT)",
            id="license-expression-with-license-classifier",
        ),
    ],
)
def test_dist_package_licenses(mocked_metadata: Mock, expected_output: str, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("pipdeptree._models.package.metadata", Mock(return_value=mocked_metadata))
    dist = DistPackage(Mock(metadata={"Name": "a"}))
    licenses_str = dist.licenses()

    assert licenses_str == expected_output


def test_dist_package_licenses_importlib_cant_find_package(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("pipdeptree._models.package.metadata", Mock(side_effect=PackageNotFoundError()))
    dist = DistPackage(Mock(metadata={"Name": "a"}))
    licenses_str = dist.licenses()

    assert licenses_str == Package.UNKNOWN_LICENSE_STR


def test_dist_package_key_pep503_normalized() -> None:
    foobar = Mock(metadata={"Name": "foo.bar"}, version="20.4.1")
    dp = DistPackage(foobar)
    assert dp.key == "foo-bar"


def test_req_package_key_pep503_normalized() -> None:
    bar_req = MagicMock(specifier=[">=4.0"])
    bar_req.name = "bar.bar-bar-bar"
    rp = ReqPackage(bar_req)
    assert rp.key == "bar-bar-bar-bar"


def test_req_package_render_as_root() -> None:
    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    bar_req = MagicMock(specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    assert rp.render_as_root(frozen=False) == "bar==4.1.0"


def test_req_package_render_as_root_with_frozen(mocker: MockerFixture) -> None:
    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    dp = DistPackage(bar)
    bar_req = MagicMock(specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dp)
    expected = "test"
    mocker.patch("pipdeptree._models.package.dist_to_frozen_repr", Mock(return_value=expected))
    assert rp.render_as_root(frozen=True) == expected


def test_req_package_render_as_branch() -> None:
    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    bar_req = MagicMock(specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    assert rp.render_as_branch(frozen=False) == "bar [required: >=4.0, installed: 4.1.0]"


def test_req_package_is_conflicting_handle_dev_versions() -> None:
    # ensure that we can handle development versions when detecting conflicts
    # see https://github.com/tox-dev/pipdeptree/issues/393
    bar = Mock(metadata={"Name": "bar"}, version="1.2.3.dev0")
    bar_req = MagicMock(specifier=SpecifierSet(">1.2.0"))
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    assert not rp.is_conflicting()


def test_req_package_as_dict() -> None:
    bar = Mock(metadata={"Name": "bar"}, version="4.1.0")
    bar_req = MagicMock(specifier=[">=4.0"])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    result = rp.as_dict()
    expected = {"key": "bar", "package_name": "bar", "installed_version": "4.1.0", "required_version": ">=4.0"}
    assert expected == result


def test_req_package_as_dict_with_no_version_spec() -> None:
    bar = Mock(key="bar", version="4.1.0")
    bar_req = MagicMock(specifier=[])
    bar_req.name = "bar"
    rp = ReqPackage(bar_req, dist=bar)
    result = rp.as_dict()
    expected = {"key": "bar", "package_name": "bar", "installed_version": "4.1.0", "required_version": "Any"}
    assert expected == result



from __future__ import annotations

import argparse
from typing import Any

import pytest

from pipdeptree._cli import EnumAction, build_parser, get_options
from pipdeptree._warning import WarningType


def test_parser_default() -> None:
    parser = build_parser()
    args = parser.parse_args([])
    assert not args.json
    assert args.output_format is None


def test_parser_j() -> None:
    parser = build_parser()
    args = parser.parse_args(["-j"])
    assert args.json
    assert args.output_format is None


def test_parser_json() -> None:
    parser = build_parser()
    args = parser.parse_args(["--json"])
    assert args.json
    assert args.output_format is None


def test_parser_json_tree() -> None:
    parser = build_parser()
    args = parser.parse_args(["--json-tree"])
    assert args.json_tree
    assert not args.json
    assert args.output_format is None


def test_parser_mermaid() -> None:
    parser = build_parser()
    args = parser.parse_args(["--mermaid"])
    assert args.mermaid
    assert not args.json
    assert args.output_format is None


def test_parser_pdf() -> None:
    parser = build_parser()
    args = parser.parse_args(["--graph-output", "pdf"])
    assert args.output_format == "pdf"
    assert not args.json


def test_parser_svg() -> None:
    parser = build_parser()
    args = parser.parse_args(["--graph-output", "svg"])
    assert args.output_format == "svg"
    assert not args.json


@pytest.mark.parametrize(
    ("should_be_error", "depth_arg", "expected_value"),
    [
        (True, ["-d", "-1"], None),
        (True, ["--depth", "string"], None),
        (False, ["-d", "0"], 0),
        (False, ["--depth", "8"], 8),
        (False, [], float("inf")),
    ],
)
def test_parser_depth(should_be_error: bool, depth_arg: list[str], expected_value: float | None) -> None:
    parser = build_parser()

    if should_be_error:
        with pytest.raises(SystemExit):
            parser.parse_args(depth_arg)
    else:
        args = parser.parse_args(depth_arg)
        assert args.depth == expected_value


def test_parser_get_options_license_and_freeze_together_not_supported(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit, match="2"):
        get_options(["--license", "--freeze"])

    out, err = capsys.readouterr()
    assert not out
    assert "cannot use --license with --freeze" in err


@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["--path", "/random/path", "--local-only"], id="path-with-local"),
        pytest.param(["--path", "/random/path", "--user-only"], id="path-with-user"),
    ],
)
def test_parser_get_options_path_with_either_local_or_user_not_supported(
    args: list[str], capsys: pytest.CaptureFixture[str]
) -> None:
    with pytest.raises(SystemExit, match="2"):
        get_options(args)

    out, err = capsys.readouterr()
    assert not out
    assert "cannot use --path with --user-only or --local-only" in err


def test_parser_get_options_exclude_dependencies_without_exclude(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit, match="2"):
        get_options(["--exclude-dependencies"])

    out, err = capsys.readouterr()
    assert not out
    assert "must use --exclude-dependencies with --exclude" in err


@pytest.mark.parametrize(("bad_type"), [None, str])
def test_enum_action_type_argument(bad_type: Any) -> None:
    with pytest.raises(TypeError, match="type must be a subclass of Enum"):
        EnumAction(["--test"], "test", type=bad_type)


def test_enum_action_default_argument_not_str() -> None:
    with pytest.raises(TypeError, match="default must be defined with a string value"):
        EnumAction(["--test"], "test", type=WarningType)


def test_enum_action_default_argument_not_a_valid_choice() -> None:
    with pytest.raises(ValueError, match="default value should be among the enum choices"):
        EnumAction(["--test"], "test", type=WarningType, default="bad-warning-type")


def test_enum_action_call_with_value() -> None:
    action = EnumAction(["--test"], "test", type=WarningType, default="silence")
    namespace = argparse.Namespace()
    action(argparse.ArgumentParser(), namespace, "suppress")
    assert getattr(namespace, "test", None) == WarningType.SUPPRESS


def test_enum_action_call_without_value() -> None:
    # ensures that we end up using the default value in case no value is specified (currently we pass nargs='?' when
    # creating the --warn option, which is why this test exists)
    action = EnumAction(["--test"], "test", type=WarningType, default="silence")
    namespace = argparse.Namespace()
    action(argparse.ArgumentParser(), namespace, None)
    assert getattr(namespace, "test", None) == WarningType.SILENCE



from __future__ import annotations

import site
import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import virtualenv

from pipdeptree.__main__ import main
from pipdeptree._discovery import get_installed_distributions

if TYPE_CHECKING:
    import pytest
    from pytest_mock import MockerFixture


def test_local_only(tmp_path: Path, mocker: MockerFixture, capfd: pytest.CaptureFixture[str]) -> None:
    venv_path = str(tmp_path / "venv")
    result = virtualenv.cli_run([venv_path, "--activators", ""])
    venv_site_packages = site.getsitepackages([venv_path])
    fake_dist = Path(venv_site_packages[0]) / "foo-1.2.5.dist-info"
    fake_dist.mkdir()
    fake_metadata = Path(fake_dist) / "METADATA"
    with fake_metadata.open("w") as f:
        f.write("Metadata-Version: 2.3\nName: foo\nVersion: 1.2.5\n")

    cmd = [str(result.creator.exe.parent / "python3"), "--local-only"]
    mocker.patch("pipdeptree._discovery.sys.prefix", venv_path)
    sys_path = sys.path.copy()
    mock_path = sys_path + venv_site_packages
    mocker.patch("pipdeptree._discovery.sys.path", mock_path)
    mocker.patch("pipdeptree._discovery.sys.argv", cmd)
    main()
    out, _ = capfd.readouterr()
    found = {i.split("==")[0] for i in out.splitlines()}
    expected = {"foo", "pip", "setuptools"}
    if sys.version_info >= (3, 12):
        expected -= {"setuptools"}  # pragma: no cover

    assert found == expected


def test_user_only(fake_dist: Path, mocker: MockerFixture, capfd: pytest.CaptureFixture[str]) -> None:
    # Make a fake user site.
    fake_user_site = str(fake_dist.parent)
    mocker.patch("pipdeptree._discovery.site.getusersitepackages", Mock(return_value=fake_user_site))

    # Add fake user site directory into a fake sys.path (normal environments will have the user site in sys.path).
    fake_sys_path = [*sys.path, fake_user_site]
    mocker.patch("pipdeptree._discovery.sys.path", fake_sys_path)

    cmd = ["", "--user-only"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)
    main()

    out, err = capfd.readouterr()
    assert not err
    found = {i.split("==")[0] for i in out.splitlines()}
    expected = {"bar"}

    assert found == expected


def test_user_only_when_in_virtual_env(
    tmp_path: Path, mocker: MockerFixture, capfd: pytest.CaptureFixture[str]
) -> None:
    # ensures that we follow `pip list` by not outputting anything when --user-only is set and pipdeptree is running in
    # a virtual environment

    # Create a virtual environment and mock sys.path to point to the venv's site packages.
    venv_path = str(tmp_path / "venv")
    virtualenv.cli_run([venv_path, "--activators", ""])
    venv_site_packages = site.getsitepackages([venv_path])
    mocker.patch("pipdeptree._discovery.sys.path", venv_site_packages)
    mocker.patch("pipdeptree._discovery.sys.prefix", venv_path)

    cmd = ["", "--user-only"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)
    main()

    out, err = capfd.readouterr()
    assert not err

    # Here we expect 1 element because print() adds a newline.
    found = out.splitlines()
    assert len(found) == 1
    assert not found[0]


def test_user_only_when_in_virtual_env_and_system_site_pkgs_enabled(
    tmp_path: Path, fake_dist: Path, mocker: MockerFixture, capfd: pytest.CaptureFixture[str]
) -> None:
    # ensures that we provide user site metadata when --user-only is set and we're in a virtual env with system site
    # packages enabled

    # Make a fake user site directory since we don't know what to expect from the real one.
    fake_user_site = str(fake_dist.parent)
    mocker.patch("pipdeptree._discovery.site.getusersitepackages", Mock(return_value=fake_user_site))

    # Create a temporary virtual environment. Add the fake user site to path (since user site packages should normally
    # be there).
    venv_path = str(tmp_path / "venv")
    virtualenv.cli_run([venv_path, "--system-site-packages", "--activators", ""])
    venv_site_packages = site.getsitepackages([venv_path])
    mock_path = sys.path + venv_site_packages + [fake_user_site]
    mocker.patch("pipdeptree._discovery.sys.path", mock_path)
    mocker.patch("pipdeptree._discovery.sys.prefix", venv_path)

    cmd = ["", "--user-only"]
    mocker.patch("pipdeptree.__main__.sys.argv", cmd)
    main()

    out, err = capfd.readouterr()
    assert not err
    found = {i.split("==")[0] for i in out.splitlines()}
    expected = {"bar"}

    assert found == expected


def test_interpreter_query_failure(mocker: MockerFixture, capfd: pytest.CaptureFixture[str]) -> None:
    cmd = ["", "--python", "i-dont-exist"]
    mocker.patch("pipdeptree._discovery.sys.argv", cmd)

    main()

    _, err = capfd.readouterr()
    assert err.startswith("Failed to query custom interpreter")


def test_duplicate_metadata(mocker: MockerFixture, capfd: pytest.CaptureFixture[str]) -> None:
    mocker.patch(
        "pipdeptree._discovery.distributions",
        Mock(
            return_value=[
                Mock(metadata={"Name": "foo"}, version="1.2.5", locate_file=Mock(return_value="/path/1")),
                Mock(metadata={"Name": "foo"}, version="5.9.0", locate_file=Mock(return_value="/path/2")),
            ]
        ),
    )

    dists = get_installed_distributions()
    assert len(dists) == 1
    # we expect it to use the first distribution found
    assert dists[0].version == "1.2.5"

    _, err = capfd.readouterr()
    expected = (
        'Warning!!! Duplicate package metadata found:\n"/path/2"\n  foo                              5.9.0       '
        '     (using 1.2.5, "/path/1")\nNOTE: This warning isn\'t a failure warning.\n---------------------------------'
        "---------------------------------------\n"
    )
    assert err == expected


def test_invalid_metadata(
    mocker: MockerFixture, capfd: pytest.CaptureFixture[str], fake_dist_with_invalid_metadata: Path
) -> None:
    fake_site_dir = str(fake_dist_with_invalid_metadata.parent)
    mocked_sys_path = [fake_site_dir]
    mocker.patch("pipdeptree._discovery.sys.path", mocked_sys_path)

    dists = get_installed_distributions()

    assert len(dists) == 0
    out, err = capfd.readouterr()
    assert not out
    assert err == (
        "Warning!!! Missing or invalid metadata found in the following site dirs:\n"
        f"{fake_site_dir}\n"
        "------------------------------------------------------------------------\n"
    )


def test_paths(fake_dist: Path) -> None:
    fake_site_dir = str(fake_dist.parent)
    mocked_path = [fake_site_dir]

    dists = get_installed_distributions(supplied_paths=mocked_path)
    assert len(dists) == 1
    assert dists[0].metadata["Name"] == "bar"


def test_paths_when_in_virtual_env(tmp_path: Path, fake_dist: Path) -> None:
    # tests to ensure that we use only the user-supplied path, not paths in the virtual env
    fake_site_dir = str(fake_dist.parent)
    mocked_path = [fake_site_dir]

    venv_path = str(tmp_path / "venv")
    s = virtualenv.cli_run([venv_path, "--activators", ""])

    dists = get_installed_distributions(interpreter=str(s.creator.exe), supplied_paths=mocked_path)
    assert len(dists) == 1
    assert dists[0].metadata["Name"] == "bar"



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pipdeptree._models import DistPackage, PackageDAG, ReqPackage


def render_text(
    tree: PackageDAG,
    *,
    max_depth: float,
    encoding: str,
    list_all: bool = True,
    include_license: bool = False,
) -> None:
    """
    Print tree as text on console.

    :param tree: the package tree
    :param max_depth: the maximum depth of the dependency tree
    :param encoding: encoding to use (use "utf-8", "utf-16", "utf-32" for unicode or anything else for legacy output)
    :param list_all: whether to list all the pkgs at the root level or only those that are the sub-dependencies
    :param include_license: provide license information
    :returns: None

    """
    nodes = get_top_level_nodes(tree, list_all=list_all)

    if encoding in {"utf-8", "utf-16", "utf-32"}:
        _render_text_with_unicode(tree, nodes, max_depth, include_license)
    else:
        _render_text_without_unicode(tree, nodes, max_depth, include_license)


def get_top_level_nodes(tree: PackageDAG, *, list_all: bool) -> list[DistPackage]:
    """
    Get a list of nodes that will appear at the first depth of the dependency tree.

    :param tree: the package tree
    :param list_all: whether to list all the pkgs at the root level or only those that are the sub-dependencies
    """
    tree = tree.sort()
    nodes = list(tree.keys())
    branch_keys = {r.key for r in chain.from_iterable(tree.values())}

    if not list_all:
        nodes = [p for p in nodes if p.key not in branch_keys]

    return nodes


def _render_text_with_unicode(
    tree: PackageDAG,
    nodes: list[DistPackage],
    max_depth: float,
    include_license: bool,  # noqa: FBT001
) -> None:
    def aux(  # noqa: PLR0913, PLR0917
        node: DistPackage | ReqPackage,
        parent: DistPackage | ReqPackage | None = None,
        indent: int = 0,
        cur_chain: list[str] | None = None,
        prefix: str = "",
        depth: int = 0,
        has_grand_parent: bool = False,  # noqa: FBT001, FBT002
        is_last_child: bool = False,  # noqa: FBT001, FBT002
        parent_is_last_child: bool = False,  # noqa: FBT001, FBT002
    ) -> list[Any]:
        cur_chain = cur_chain or []
        node_str = node.render(parent, frozen=False)
        next_prefix = ""
        next_indent = indent + 2

        if parent:
            bullet = "├── "
            if is_last_child:
                bullet = "└── "

            if has_grand_parent:
                next_indent -= 1
                if parent_is_last_child:
                    prefix += " " * (indent + 1 - depth)
                else:
                    prefix += "│" + " " * (indent - depth)
                # Without this extra space, bullets will point to the space just before the project name
                prefix += " "
            next_prefix = prefix
            node_str = prefix + bullet + node_str
        elif include_license:
            node_str += " " + node.licenses()

        result = [node_str]

        children = tree.get_children(node.key)
        children_strings = [
            aux(
                c,
                node,
                indent=next_indent,
                cur_chain=[*cur_chain, c.project_name],
                prefix=next_prefix,
                depth=depth + 1,
                has_grand_parent=parent is not None,
                is_last_child=c is children[-1],
                parent_is_last_child=is_last_child,
            )
            for c in children
            if c.project_name not in cur_chain and depth + 1 <= max_depth
        ]

        result += list(chain.from_iterable(children_strings))
        return result

    lines = chain.from_iterable([aux(p) for p in nodes])
    print("\n".join(lines))  # noqa: T201


def _render_text_without_unicode(
    tree: PackageDAG,
    nodes: list[DistPackage],
    max_depth: float,
    include_license: bool,  # noqa: FBT001
) -> None:
    def aux(
        node: DistPackage | ReqPackage,
        parent: DistPackage | ReqPackage | None = None,
        indent: int = 0,
        cur_chain: list[str] | None = None,
        depth: int = 0,
    ) -> list[Any]:
        cur_chain = cur_chain or []
        node_str = node.render(parent, frozen=False)
        if parent:
            prefix = " " * indent + "- "
            node_str = prefix + node_str
        elif include_license:
            node_str += " " + node.licenses()
        result = [node_str]
        children = [
            aux(c, node, indent=indent + 2, cur_chain=[*cur_chain, c.project_name], depth=depth + 1)
            for c in tree.get_children(node.key)
            if c.project_name not in cur_chain and depth + 1 <= max_depth
        ]
        result += list(chain.from_iterable(children))
        return result

    lines = chain.from_iterable([aux(p) for p in nodes])
    print("\n".join(lines))  # noqa: T201


__all__ = ["get_top_level_nodes", "render_text"]



from __future__ import annotations

import sys
from collections import defaultdict
from typing import TYPE_CHECKING

from pipdeptree._warning import get_warning_printer

if TYPE_CHECKING:
    from pipdeptree._models.package import Package

    from ._models import DistPackage, PackageDAG, ReqPackage


def validate(tree: PackageDAG) -> None:
    # Before any reversing or filtering, show warnings to console, about possibly conflicting or cyclic deps if found
    # and warnings are enabled (i.e. only if output is to be printed to console)
    warning_printer = get_warning_printer()
    if warning_printer.should_warn():
        conflicts = conflicting_deps(tree)
        if conflicts:
            warning_printer.print_multi_line(
                "Possibly conflicting dependencies found", lambda: render_conflicts_text(conflicts)
            )

        cycles = cyclic_deps(tree)
        if cycles:
            warning_printer.print_multi_line("Cyclic dependencies found", lambda: render_cycles_text(cycles))


def conflicting_deps(tree: PackageDAG) -> dict[DistPackage, list[ReqPackage]]:
    """
    Return dependencies which are not present or conflict with the requirements of other packages.

    e.g. will warn if pkg1 requires pkg2==2.0 and pkg2==1.0 is installed

    :param tree: the requirements tree (dict)
    :returns: dict of DistPackage -> list of unsatisfied/unknown ReqPackage
    :rtype: dict

    """
    conflicting = defaultdict(list)
    for package, requires in tree.items():
        for req in requires:
            if req.is_conflicting():
                conflicting[package].append(req)
    return conflicting


def render_conflicts_text(conflicts: dict[DistPackage, list[ReqPackage]]) -> None:
    # Enforce alphabetical order when listing conflicts
    pkgs = sorted(conflicts.keys())
    for p in pkgs:
        pkg = p.render_as_root(frozen=False)
        print(f"* {pkg}", file=sys.stderr)  # noqa: T201
        for req in conflicts[p]:
            req_str = req.render_as_branch(frozen=False)
            print(f" - {req_str}", file=sys.stderr)  # noqa: T201


def cyclic_deps(tree: PackageDAG) -> list[list[Package]]:
    """
    Return cyclic dependencies as list of lists.

    :param  tree: package tree/dag
    :returns: list of lists, where each list represents a cycle

    """

    def dfs(root: DistPackage, current: Package, visited: set[str], cdeps: list[Package]) -> bool:
        if current.key not in visited:
            visited.add(current.key)
            current_dist = tree.get_node_as_parent(current.key)
            if not current_dist:
                return False

            reqs = tree.get(current_dist)
            if not reqs:
                return False

            for req in reqs:
                if dfs(root, req, visited, cdeps):
                    cdeps.append(current)
                    return True
        elif current.key == root.key:
            cdeps.append(current)
            return True
        return False

    cycles: list[list[Package]] = []

    for p in tree:
        cdeps: list[Package] = []
        visited: set[str] = set()
        if dfs(p, p, visited, cdeps):
            cdeps.reverse()
            cycles.append(cdeps)

    return cycles


def render_cycles_text(cycles: list[list[Package]]) -> None:
    # List in alphabetical order the dependency that caused the cycle (i.e. the second-to-last Package element)
    cycles = sorted(cycles, key=lambda c: c[len(c) - 2].key)
    for cycle in cycles:
        print("*", end=" ", file=sys.stderr)  # noqa: T201

        size = len(cycle) - 1
        for idx, pkg in enumerate(cycle):
            if idx == size:
                print(f"{pkg.project_name}", end="", file=sys.stderr)  # noqa: T201
            else:
                print(f"{pkg.project_name} =>", end=" ", file=sys.stderr)  # noqa: T201
        print(file=sys.stderr)  # noqa: T201


__all__ = [
    "validate",
]
